


#include <stdint.h>
#include <stdbool.h>
#include <pru_cfg.h>
#include "resource_table_empty.h"
#include "sharedDataStruct.h"

// Reference for shared RAM:
// https://markayoder.github.io/PRUCookbook/05blocks/blocks.html#_controlling_the_pwm_frequency
// GPIO Configuration
// ----------------------------------------------------------

volatile register uint32_t __R30; // output GPIO register
volatile register uint32_t __R31; // input GPIO register

/**
 * Set this to values greater than 0 to enable PRU debugging mode. 
 * You must also set DEBUG_PRU in tempSensor.c to values greater than 1 to 
 * properly debug the PRU. 
 * In debugging mode, values are initialized to known values and printed 
 * to the console once; the PRU then executes a looping command on the 
 * debugging LED to verify wiring setup. 
*/
#define DEBUG_PRU 0

// Use pru0_rpu_r30_5 as an output (bit 5 = 0b0010 0000)
#define LED_MASK (1 << 5)
// Use pru0_pru_r31_3 as a button (bit 3 = 0b0000 1000)
#define BUTTON_MASK (1 << 3)
//use pru0_r30_0 as a GPIO output (bit 0 = 0b0000 0001)
//pin 9.31 (linux 110)
#define OUTPUT_PIN_MASK (1 << 0)
//use pru0_r31_1 as a GPIO input (bit 1 = 0b0000 0010)
//pin 9.29 (linux 111)
#define INPUT_PIN_MASK (1 << 1)

#define DELAY_1_MS (200000)
#define DELAY_1_US (200)
#define DELAY_LOGICAL_0 (DELAY_1_US * 40)
#define DELAY_LOCICAL_1 (DELAY_1_US * 75)

// Shared Memory Configuration
// -----------------------------------------------------------
#define THIS_PRU_DRAM 0x00000 // Address of DRAM
#define OFFSET 0x200 // Skip 0x100 for Stack, 0x100 for Heap (from makefile)
#define THIS_PRU_DRAM_USABLE (THIS_PRU_DRAM + OFFSET)

// This works for both PRU0 and PRU1 as both map their own memory to 0x0000000
volatile sharedMemStruct_t *pSharedMemStruct = (volatile void*)THIS_PRU_DRAM_USABLE;

#if DEBUG_PRU > 0
    static void debugPRU(void);
#endif
static void waitTilTimeToRead(void);
static void readFromSensor(void);
static int readFromGPIO(void);
static void writeToGPIO(bool val);

static void clearSignalData(void);

//for use in signalling 
static void turnOnLED(void);
static void turnOffLED(void);

void main(void)
{
    #if DEBUG_PRU > 0
        debugPRU();
        return;
    #endif

    // Initialize:
    turnOffLED();
    pSharedMemStruct->startReadingFromSensor = false;
    pSharedMemStruct->interruptReading = false;
    pSharedMemStruct->doneReading = false;
    pSharedMemStruct->bitsRead = 0;
    writeToGPIO(1);
    // for(int i = 0; i < READING_LENGTH; i++) {
    //     pSharedMemStruct->readValues[i] = false;
    // }

    while (true) {
        waitTilTimeToRead();
        //turnOnLED();
        readFromSensor();
        //turnOffLED();
    }
}

static void waitTilTimeToRead(void)
{
    while(!pSharedMemStruct->startReadingFromSensor) {
        //do nothing and wait
    }
    pSharedMemStruct->startReadingFromSensor = false;
}

static void readFromSensor(void)
{
    clearSignalData();

    //pull gpio low, then high 
    writeToGPIO(0);
    __delay_cycles(10*DELAY_1_MS);
    writeToGPIO(1);
    __delay_cycles(20*DELAY_1_US);

    //wait for response 
    __delay_cycles(10*DELAY_1_US); //account for finite falling edge time 
    while(readFromGPIO() == 0) {
        //do nothing and wait 
        //response is pulling low 
        if(pSharedMemStruct->interruptReading) { return; }
    }
    while(readFromGPIO() == 1) {
        //do nothing and wait 
        //response is pulling high 
        if(pSharedMemStruct->interruptReading) { return; }
    }

    //start transmitting actual data 
    for(int i = 0; i < READING_LENGTH; i++) {
        unsigned char bitRead = 0;
        
        while(readFromGPIO() == 0) {
            //sensor pulls low 
            //do nothing and wait
            if(pSharedMemStruct->interruptReading) { return; }
        }
        //now we've pulled high 
        turnOnLED();
        //see if its a short or long high 
        __delay_cycles(DELAY_LOGICAL_0);
        if(readFromGPIO() == 1) {
            //sensor is still pulling high after the duration of a logical 0, so its a 1
            bitRead = 1;
            while(readFromGPIO()) {
                //let it time out 
                if(pSharedMemStruct->interruptReading) { return; }
            }
            __delay_cycles(5 * DELAY_1_US); //account for finite falling edge time 
        }
        else {
            bitRead = 0;
        }
        turnOffLED();

        pSharedMemStruct->readValues[i] = bitRead;
        pSharedMemStruct->bitsRead++;
    }
    pSharedMemStruct->doneReading = true;
}

static int readFromGPIO(void)
{
    return ((__R31 & INPUT_PIN_MASK) != 0);
}

static void writeToGPIO(bool val)
{
    if(val) {
        __R30 |= OUTPUT_PIN_MASK;
    }
    else {
        __R30 &= ~OUTPUT_PIN_MASK;
    }
}

static void clearSignalData(void)
{
    pSharedMemStruct->bitsRead = 0;
    pSharedMemStruct->startReadingFromSensor = false;
    pSharedMemStruct->doneReading = false;
    pSharedMemStruct->interruptReading = false;
    for(int i = 0; i < READING_LENGTH; i++) {
        pSharedMemStruct->readValues[i] = false;
    }
}

static void turnOnLED(void)
{
    __R30 |= LED_MASK;  
}

static void turnOffLED(void) 
{
    __R30 &= ~LED_MASK;
}

#if DEBUG_PRU > 0
    static void debugPRU(void)
    {
        pSharedMemStruct->startReadingFromSensor = true;
        pSharedMemStruct->interruptReading = false;
        for(int i = 0; i < READING_LENGTH; i++) {
            pSharedMemStruct->readValues[i] = (i % 2 ? 1 : 0);
        }

        while(true) {
            /**
             * Uncomment these to debug individual components of the system, 
             * Functionality confirmed increases as tests go from bottom upwards; 
             * i.e. the lowest test is the easiest to confirm but is least relevant to 
             * the project; the highest is hardest but is most useful to the project. 
            */

            //read from input pin mask, and turn on LED when button is pressed.
            //LED should be connected to the DATA WRITE pin
            if(readFromGPIO()) {
                writeToGPIO(1);
            }
            else {
                writeToGPIO(0);
            }

            //read from input pin mask, and turn on LED when button is pressed. 
            //LED should be connected to the LED WRITE pin
            /*if(readFromGPIO()) {
                turnOnLED();
            }
            else {
                turnOffLED();
            }*/
            
            //makes LED flash at 1Hz 
            //LED should be connected to the DATA WRITE pin
            /*writeToGPIO(1);
            __delay_cycles(DELAY_1_MS * 500);
            writeToGPIO(0);
            __delay_cycles(DELAY_1_MS * 500);*/
           
           //makes LED flash at 1Hz 
           //LED should be connected to the LED WRITE pin
           /*turnOnLED();
           __delay_cycles(DELAY_1_MS * 500);
           turnOffLED();
           __delay_cycles(DELAY_1_MS * 500);*/
        }
    }
#endif
