#include "tempSensor.h"
#include "common.h"
#include "project-pru/sharedDataStruct.h"

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <pthread.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>  
#include <sys/mman.h>

/**
 * set this to values greater than 0 to enable debugging mode
 * debugging mode will log information to the console such as: 
 *     -values read by the sensor in real time 
 *     -actions being performed, such as initialization processes 
*/
#define DEBUG 0
/**
 * Set this to values greater than 0 to enable PRU debugging mode. 
 * You must also set DEBUG_PRU in pruMain.c to values greater than 1 to 
 * properly debug the PRU. 
 * In debugging mode, values are initialized to known values and printed 
 * to the console once; the PRU then executes a looping command on the 
 * debugging LED to verify wiring setup. 
*/
#define DEBUG_PRU 0

// General PRU Memomry Sharing Routine
// ----------------------------------------------------------------
#define PRU_ADDR 0x4A300000 // Start of PRU memory Page 184 am335x TRM
#define PRU_LEN 0x80000 // Length of PRU memory
#define PRU0_DRAM 0x00000 // Offset to DRAM
#define PRU1_DRAM 0x02000
#define PRU_SHAREDMEM 0x10000 // Offset to shared memory
#define PRU_MEM_RESERVED 0x200 // Amount used by stack and heap

// Convert base address to each memory section
#define PRU0_MEM_FROM_BASE(base) ( (base) + PRU0_DRAM + PRU_MEM_RESERVED)
#define PRU1_MEM_FROM_BASE(base) ( (base) + PRU1_DRAM + PRU_MEM_RESERVED)
#define PRUSHARED_MEM_FROM_BASE(base) ( (base) + PRU_SHAREDMEM)

//pthread components
static pthread_t sensor;
static pthread_mutex_t accessMutex = PTHREAD_MUTEX_INITIALIZER;
static bool cleanup = false;

//data to share through gets/sets
static float currentTemp = 0;
static float currentPercentHumidity = 0;
static bool hasReadValues = false;

//sensor static methods 
static void* sensorThread(void* useless);
static void initializeSensorGPIO(void);
static bool beginReadingFromSensor(void);

//pru components 
static volatile sharedMemStruct_t* pruData;
static volatile void* pPruBase;

//pru methods 
static void initializePRuSharedMemory(void);
static volatile void* getPruMmapAddr(void);
static void freePruMmapAddr(volatile void* pPruBase);
static void resetPruData(void);
#if DEBUG_PRU > 0
    static void debugPRU(void);
#endif

void tempSensor_startSampling(void)
{
    initializeSensorGPIO();
    initializePRuSharedMemory();
    #if DEBUG_PRU > 0
        debugPRU();
        return;
    #endif
    pthread_create(&sensor, NULL, &sensorThread, NULL);
}

void tempSensor_stopSampling(void)
{
    cleanup = true; 
    pthread_join(sensor, NULL);
    freePruMmapAddr(pPruBase);
}

static void* sensorThread(void* useless) 
{
    //must delay until the sensor has stabilized 
    sleepForMs(1100);
    while(!cleanup) {
        beginReadingFromSensor();
        pthread_mutex_lock(&accessMutex);
        {
            hasReadValues = true;
        }
        pthread_mutex_unlock(&accessMutex);
        sleepForMs(8000);
    }
    return NULL;
}

static void initializeSensorGPIO(void)
{
    //PRU
    runCommand("config-pin p9_27 pruout");
    runCommand("config-pin p9_28 pruin");
    runCommand("config-pin p9_29 pruin");
    runCommand("config-pin p9_31 pruout");
    sleepForMs(500);
    #if DEBUG > 0
        printf("Initialized pru pins!\n");
    #endif
}

static bool beginReadingFromSensor(void)
{
    #if DEBUG > 0
        printf("Starting to read from sensor...\n");
    #endif

    long long startTime = getTimeInMs();
    pruData->startReadingFromSensor = true;

    //wait til it sends the done signal or takes too long 
    bool hadToInterrupt = false;
    while(!pruData->doneReading) {
        sleepForMs(100);
        if(getTimeInMs() - startTime > 5000) {
            hadToInterrupt = true;
            pruData->interruptReading = true;
            break;
        }
    }
    if(hadToInterrupt) {
        printf("Error in reading from pru: pru timed out.\n");
        return false;
    }

    #if DEBUG > 1
        printf("Sensor read values: \n");
        int lineBreak = 0;
        for(int i = 0; i < READING_LENGTH; i++) {
            printf("%d", pruData->readValues[i]);
            lineBreak++; 
            if(lineBreak == 4) { printf(" "); lineBreak = 0; }
        }
        printf("\n");
    #endif

    //convert data to numbers 
    unsigned char tempHigher = 0; 
    unsigned char tempLower = 0;
    unsigned char humidityHigher = 0;
    unsigned char humidityLower = 0; 
    unsigned char checksum = 0;
    for(int i = 0; i < 8; i++) {
        //shift existing bits left for all things 
        tempHigher = tempHigher << 1;
        tempLower = tempLower << 1;
        humidityHigher = humidityHigher << 1;
        humidityLower = humidityLower << 1;
        checksum = checksum << 1;

        //OR in new bits 
        humidityHigher |= pruData->readValues[0 + i];
        humidityLower |= pruData->readValues[8 + i];
        tempHigher |= pruData->readValues[16 + i];
        tempLower |= pruData->readValues[24 + i];
        checksum |= pruData->readValues[32 + i];
    }
    if(!(checksum == ((tempHigher + tempLower + humidityHigher + humidityLower) % 0x100))) {
        printf("Error in transmission: checksum does not match elements.\n");
        #if DEBUG > 0
            printf("Checksum: %d\n", checksum);
            printf("Sum of elements: %d\n", (humidityHigher + humidityLower + tempHigher + tempLower) % 0x100);
            printf("Array elements: \n");
            printf("%02x %02x %02x %02x %02x\n", humidityHigher, humidityLower, tempHigher, tempLower, checksum);
        #endif
        return false;
    }

    #if DEBUG > 1
        printf("You've read values from the sensor successfully!\n");
    #endif

    bool tempIsNegative = false;
    //do check for if temp is negative
    if(tempHigher & 0b10000000) {
        #if DEBUG > 1
            printf("Negative temp detected!\n");
        #endif
        tempHigher &= 0b1111111;
        tempIsNegative = true;
    }

    //process the read values 
    unsigned totalTemp = tempHigher;
    totalTemp = totalTemp << 8;
    totalTemp |= tempLower;
    unsigned totalHumidity = humidityHigher;
    totalHumidity = totalHumidity << 8;
    totalHumidity |= humidityLower;

    //copy into floats and prepare to store into static globals 
    float newTemp = totalTemp;
    if(tempIsNegative) {
        newTemp *= -1;
    }
    float newHumidity = totalHumidity;
    newTemp /= 10;
    newHumidity /= 10;

    #if DEBUG > 0
        printf("Read values: temp = %f, hum = %f\n", newTemp, newHumidity);
    #endif

    pthread_mutex_lock(&accessMutex);
    {
        currentTemp = newTemp;
        currentPercentHumidity = newHumidity;
    }
    pthread_mutex_unlock(&accessMutex);
    return true;
}

float tempSensor_getCurrentTemp(void)
{
    return currentTemp;
}

float tempSensor_getCurrentPercentHumidity(void)
{
    return currentPercentHumidity;
}

bool tempSensor_readyToReadValues(void)
{
    return hasReadValues;
}

static void initializePRuSharedMemory(void) 
{
    #if DEBUG > 0
        printf("initializing pru shared mem...\n");
    #endif
    // Get access to shared memory for my uses
    pPruBase = getPruMmapAddr();
    pruData = PRU0_MEM_FROM_BASE(pPruBase);

    resetPruData();

    #if DEBUG > 0
        printf("initialized pru memory!\n");
    #endif
}

// Return the address of the PRU's base memory
static volatile void* getPruMmapAddr(void) {
    int fd = open("/dev/mem", O_RDWR | O_SYNC);
    if (fd == -1) {
        perror("ERROR: could not open /dev/mem");
        exit(EXIT_FAILURE);
    }
    // Points to start of PRU memory.
    volatile void* pruBase = mmap(0, PRU_LEN, PROT_READ | PROT_WRITE,
    MAP_SHARED, fd, PRU_ADDR);
    if (pruBase == MAP_FAILED) {
        perror("ERROR: could not map memory");
        exit(EXIT_FAILURE);
    }
    close(fd);
    return pruBase;
}

static void freePruMmapAddr(volatile void* pruBase) {
    if (munmap((void*) pruBase, PRU_LEN)) {
        perror("PRU munmap failed");
        exit(EXIT_FAILURE);
    }
}

static void resetPruData(void)
{
    pruData->startReadingFromSensor = false;
    pruData->bitsRead = 0;
    pruData->doneReading = false;
    pruData->interruptReading = false;
    for(int i = 0; i < READING_LENGTH; i++) {
        pruData->readValues[i] = false;
    }
} 

#if DEBUG_PRU > 0
static void debugPRU(void) 
{
    printf("NOW DEBUGGING PRU...\n");
    printf("startReadingFromSensor: %d\n", pruData->startReadingFromSensor);
    printf("Read values: \n");
    for(int i = 0; i < READING_LENGTH; i++) {
        printf("%d", pruData->readValues[i]);
        if(i && !(i % 8)) {
            printf("\n");
        }
    }
    printf("\n");
    printf("doneReading: %d\n", pruData->doneReading);
    printf("interruptReading: %d\n", pruData->interruptReading);
    printf("bits read: %d\n", pruData->bitsRead);
}
#endif 